#include <stdio.h>
#include <errno.h>
#include <math.h>
#ifdef MPI
#include <mpi.h>
#endif

#include "DEFS.h"
#include "PATH.h"
#include "PROB.h"
#include "MEM.h"
#include "PROCESS.h"

#ifdef MPI
  extern int MyProcessorNumber, NumberofProcessors;
#endif

static PROCESS *SetParams
                        (
                        int ,
                        int ,
                        REAL *,
                        int *,
                        REAL **
                        );

static void SetCouplings
                       (
                       PROCESS *,
                       int ,
                       int *,
                       REAL **
                       );

static void SetVariances
                       (
                       PROCESS *,
                       int ,
                       REAL *
                       );

static void SetProbabilityFunction
                                 (
                                 int ,
                                 PROCESS *
                                 );



PROCESS *SetProcessParams
                        (
                        int NumTimeSteps, 
                        int NumRates,
                        PATH *Disturbance,
                        char *DISTURBANCE_description_file
                        )
{
  FILE *dfile;
  int err = 0;
  int DisturbanceType;
  PROCESS *ProcessParams = NULL;
  REAL **Couplings = NULL;
  REAL *Variances = NULL;
  REAL *InitValues = NULL;
  int *NumOfCouplings = NULL;
  int ir,j,jj,iir,size,Ncouplings;
  float vir;
#ifdef MPI
  if(MyProcessorNumber == MASTER) {
#endif MPI

  if((dfile = fopen(DISTURBANCE_description_file,"r")) == NULL) {
        fprintf(stderr, "SetDisturbanceParams(): can't open description file\n");
        exit(1);
  }
  getline(dfile);
  fscanf(dfile,"%d\n",&DisturbanceType);
#ifdef MPI
  }
#endif MPI

#ifdef MPI
  MPI_Bcast(&DisturbanceType, 1, MPI_INT, 0, MPI_COMM_WORLD);
  MPI_Barrier(MPI_COMM_WORLD);
#endif MPI


  size=(NumRates+1)*sizeof(REAL);
  Variances = (REAL *) mymalloc(size,"SetDisturbanceParams():Variances\n");
  InitValues = (REAL *) mymalloc(size,"SetDisturbanceParams():InitValues\n");
  NumOfCouplings = (int *) mymalloc(size,"SetDisturbanceParams():NumOfCouplings\n");

#ifdef MPI
  if(MyProcessorNumber == MASTER) {
#endif MPI

  getline(dfile);
  for(ir=1;ir<=NumRates;ir++){
#ifdef DOUBLE_PRECISION
     fscanf(dfile," %d %lf\n",&iir,Variances+ir);
#else
     fscanf(dfile," %d %f\n",&iir,Variances+ir);
#endif
  }

#ifdef MPI
  }
#endif MPI
#ifdef MPI
  MPI_Bcast(Variances, NumRates+1, MPI_REAL, 0, MPI_COMM_WORLD);
#endif MPI
  size=(NumRates+1)*sizeof(REAL *);
  Couplings = (REAL **) mymalloc(size,"SetDisturbanceParams():Couplings\n");
  Couplings[0]=(REAL *) mymalloc(sizeof(REAL),"SetDisturbanceParams():Couplings\n");

#ifdef MPI
  if(MyProcessorNumber == MASTER) {
#endif MPI

  getline(dfile);
  for(ir=1;ir<=NumRates;ir++){
     fscanf(dfile," %d %d\n",&iir,&Ncouplings);
#ifdef MPI
     MPI_Bcast(&Ncouplings, 1, MPI_INT, 0, MPI_COMM_WORLD);
#endif
/* Scott ---> help with this */
     size = (Ncouplings+1)*sizeof(REAL);
     NumOfCouplings[ir] = Ncouplings;
     Couplings[ir]=(REAL *) mymalloc(size,"SetDisturbanceParams():Couplings\n");
     if(Ncouplings > 0) {
       for(j=1;j<=Ncouplings;j++){
#ifdef DOUBLE_PRECISION
          fscanf(dfile," %d %d %lf\n",&iir,&jj,Couplings[ir]+j);
#else
          fscanf(dfile," %d %d %f\n",&iir,&jj,Couplings[ir]+j);
#endif
       }
#ifdef MPI
            MPI_Bcast(Couplings[ir], Ncouplings+1, MPI_REAL, 0, MPI_COMM_WORLD);
#endif
     }
  }
#ifdef MPI
  } else {

    for(ir=1;ir<=NumRates;ir++){
      MPI_Bcast(&Ncouplings, 1, MPI_INT, 0, MPI_COMM_WORLD);

      size = (Ncouplings+1)*sizeof(REAL);
      NumOfCouplings[ir] = Ncouplings;
      Couplings[ir]=(REAL *) mymalloc(size,"SetDisturbanceParams():Couplings\n");

      if(Ncouplings > 0)
        MPI_Bcast(Couplings[ir], Ncouplings+1, MPI_REAL, 0, MPI_COMM_WORLD);
      }


  }
#endif


#ifdef MPI
  if(MyProcessorNumber == MASTER) {
#endif MPI

  getline(dfile);
  for(ir=1;ir<=NumRates;ir++){
#ifdef DOUBLE_PRECISION
     fscanf(dfile," %d %lf\n",&iir,InitValues+ir);
#else
     fscanf(dfile," %d %f\n",&iir,InitValues+ir);
#endif
  }
#ifdef MPI
  }
#endif
#ifdef MPI
  MPI_Bcast(InitValues, NumRates+1, MPI_REAL, 0, MPI_COMM_WORLD);
#endif


  SetCurrentTimetoInitial(Disturbance);
  SetCurrentTimeSlice
                    (
                    Disturbance,
                    InitValues
                    );

  ProcessParams = SetParams
                          ( 
                          DisturbanceType, 
                          NumRates,
                          Variances, 
                          NumOfCouplings, 
                          Couplings
                          );

  M_Free((void **)Couplings, NumRates);
  V_Free((void *)NumOfCouplings);
  V_Free((void *)Variances);
  V_Free((void *)InitValues);
  
  return ProcessParams;
}



static PROCESS *SetParams
                        (
                        int ProcessID,
                        int NumRates, 
                        REAL *Variances, 
                        int *NumOfCouplings,
                        REAL **Couplings
                        )
{
  int i;
  int err = 0;
  PROCESS *ProcessParams = NULL;

  if((ProcessParams=
     (PROCESS *)mymalloc(sizeof(PROCESS),"SetParams():ProcessParams\n")) == NULL) {
   fprintf(stderr,"SetProcessParam():cannot allocate buffer for ProcessParams\n");
   exit(1);
  }

  ProcessParams->Dimension = NumRates;

  ProcessParams->ProcessID = ProcessID;
  SetProbabilityFunction(ProcessID, ProcessParams);

  if((ProcessParams->Variance=(REAL *)
        malloc((NumRates+1)*sizeof(REAL))) == NULL) {
        fprintf(stderr,"SetProcessParams():cannot allocate buffer for variance \n");
        exit(1);
  }
  SetVariances
             (
             ProcessParams, 
             NumRates, 
             Variances
             );

  if((ProcessParams->RateCouplings=(REAL **)
        malloc((NumRates+1)*sizeof(REAL *))) == NULL) {
        fprintf(stderr,"SetProcessParams():cannot allocate buffer for couplings \n");
        exit(1);
  }

  if((ProcessParams->CouplingsMap=(int *)
        malloc((NumRates+1)*sizeof(int))) == NULL) {
        fprintf(stderr,"SetProcessParams():cannot allocate buffer for couplings map \n");
        exit(1);
  }

  for(i=0;i<=NumRates;i++) {
    if((ProcessParams->RateCouplings[i]=(REAL *)
        malloc((NumOfCouplings[i]+1)*sizeof(REAL))) == NULL) {
        fprintf(stderr,"SetProcessParams():can't allocate buffer for coupling # %d\n",i);
        exit(1);
    }
  }

  SetCouplings
             (
             ProcessParams, 
             NumRates, 
             NumOfCouplings, 
             Couplings
             );

  return ProcessParams;
}



int GetCurrentRateCouplings
                          (
                          PROCESS *ProcessParams,
                          PATH *PathParams,
                          int NumCouplings,
                          REAL *p
                          )
{

  int ir;
  int NumRates         = ProcessParams->Dimension;
  int CurrentRate      = GetCurrentRateIndex(PathParams);

  if(p != NULL) {
     for(ir=0; ir <= NumCouplings; ir ++)
        *(p+ir) = *(ProcessParams->RateCouplings[CurrentRate] + ir);
  } else {
        fprintf(stderr, "GetCurrentRateCouplings(): Invalid pointer argument\n");
        exit(1);
  }

  return DONE;
}




int ProcessSymmetry
                  (
                  PROCESS *Params, 
                  int TypeOfSymmetry
                  )
{
 return ((Params->Symmetry) & (TypeOfSymmetry));
}



int *GetCouplingsMap
                   (
                   PROCESS *ProcessParams, 
                   int NumRates 
                   )
{

  int ir;
  int *p = NULL;
  int size = (NumRates + 1)*sizeof(int);

  p = (int *) mymalloc(size,"GetCouplingsMap()\n");
  if(p != NULL) {
    p[0] = 0;
    for(ir=1; ir <= NumRates; ir ++) *(p+ir) = *(ProcessParams->CouplingsMap + ir);
  } else {
     fprintf(stderr, "GetCouplingsMap(): Invalid pointer argument\n");
     exit(1);
  }
  return p;
}



REAL *GetVariances
                  (
                  PROCESS *ProcessParams, 
                  int NumRates 
                  )
{

  int ir;
  REAL *p = NULL;
  int size = (NumRates + 1)*sizeof(REAL);

  p = (REAL *) mymalloc(size,"GetVariances()\n");
  if(p != NULL) {
    p[0] = 0;
    for(ir=1; ir <= NumRates; ir ++) *(p+ir) = *(ProcessParams->Variance + ir);
  } else {
     fprintf(stderr, "GetVariances(): Invalid pointer argument\n");
     exit(1);
  }
  return p;
}



REAL **GetAllRateCouplings
                          (
                          PROCESS *ProcessParams
                          )
{
  int ir;
  int jr;
  int size;
  int NumRates  = ProcessParams->Dimension;
  int *NumCouplings = NULL;
  REAL **p = NULL;

  size = (NumRates + 1)*sizeof(REAL *);
  p = (REAL **) mymalloc(size,"GetAllRateCouplings(): AllCouplings\n");

  NumCouplings =  GetCouplingsMap(ProcessParams,NumRates);
  for(ir=0;ir<=NumRates;ir++) {
      size = (NumCouplings[ir] + 1)*sizeof(REAL);
      p[ir] = (REAL *) mymalloc(size,"MetropolisUpdate(): AllCouplings\n");
   }

  if(p != NULL) {
     for(ir=0; ir<= NumRates; ir++)
        for(jr=0; jr <= NumCouplings[ir]; jr ++) {
           p[ir][jr]  = *(ProcessParams->RateCouplings[ir] + jr);
        }
  } else {
     fprintf(stderr, "GetAllRateCouplings(): Invalid pointer argument\n");
     exit(1);
  }

  V_Free((void *) NumCouplings);
  return p;
}



REAL *GetVectorOfCouplings
                         (
                         PROCESS *ProcessParams
                         )
{
  int ir;
  int jr;
  int size;
  int NumRates  = ProcessParams->Dimension;
  int *NumCouplings = NULL;
  REAL *p = NULL;
  int NumParams;
  int index;

  NumCouplings =  GetCouplingsMap(ProcessParams,NumRates);

  NumParams = 0;
  for(ir=1;ir<=NumRates;ir++) {
      NumParams += NumCouplings[ir] ;
   }
  size = (NumParams + 1)*sizeof(REAL *);
  p = (REAL *) mymalloc(size,"GetVectorOfCouplings(): Params\n");

  if(p != NULL) {
     p[0] = 0;
     index = 0;
     for(ir=1; ir<= NumRates; ir++)
        for(jr=1; jr <= NumCouplings[ir]; jr ++)
           index++;
           *(p+index) = *(ProcessParams->RateCouplings[ir] + jr);
  } else {
     fprintf(stderr, "GetVector OfCouplings(): Invalid pointer argument\n");
     exit(1);
  }

  if(index > NumParams) {
     fprintf(stderr, "GetVector OfCouplings(): array size mismatch!\n");
     exit(1);
  }

  V_Free((void *) NumCouplings);
  return p;
}




int HowManyParams
                (
                PROCESS *ProcessParams
                )
{
  int ir;
  int NumRates  = ProcessParams->Dimension;
  int *NumCouplings = NULL;
  int NumParams;

  NumCouplings =  GetCouplingsMap
                                (
                                ProcessParams,
                                NumRates
                                );

  NumParams = 0;
  for(ir=1;ir<=NumRates;ir++) {
      NumParams += NumCouplings[ir] ;
  }
  return NumParams;
}



REAL GetVariance
                (
                PROCESS *ProcessParams,
                PATH *PathParams
                )
{
 int ir = GetCurrentRateIndex(PathParams);
 return *(ProcessParams->Variance + ir);
}



int GetProcessID
               (
               PROCESS *ProcessParams
               )
{
  int ev;
  ev = ProcessParams->ProcessID;
  return ev;
}



static void SetCouplings
                       ( 
                       PROCESS *ProcessParams, 
                       int NumRates, 
                       int *NumOfCouplings, 
                       REAL **Couplings
                       )
{

  int i,j;

  *(ProcessParams->CouplingsMap+0) = 0;

  for(i=1;i<=NumRates;i++) {

      *(ProcessParams->CouplingsMap+i) = NumOfCouplings[i];

      for(j=1;(NumOfCouplings[i] > 0) && (j<=NumOfCouplings[i]);j++)
               *(ProcessParams->RateCouplings[i]+j)= *(Couplings[i] + j);
  }

}


static void SetVariances
                       (
                       PROCESS *ProcessParams, 
                       int NumRates, 
                       REAL *Variances
                       )
{
  int i;

  for(i=1;i<=NumRates;i++) {
         *(ProcessParams->Variance+i)=Variances[i];
  }
}



ProbFunc *GetProbabilityFunction
                               (
                               PROCESS *Params
                               )
{
 return Params->Probability;
}



static void SetProbabilityFunction
                                 (
                                 int probID, 
                                 PROCESS *Params
                                 )
{
  ProbFunc *pf;

  Params->Symmetry = NONE;
#ifdef BUNDLE_AR_WITH_PROB
  switch(probID)  {
  case 1:
          pf = GaussianT;
          Params->Probability = pf;
          Params->Symmetry |= (TRANSLATION | REFLEXION);
          break;
  case 2:
          pf = Gaussian;
          Params->Probability = pf;
          Params->Symmetry |= REFLEXION ;
          break;
  case 3:
          pf = Student3T;
          Params->Probability = pf;
          Params->Symmetry |= (TRANSLATION | REFLEXION) ;
          break;
  case 4:
          pf = GeneralizedCauchy;
          Params->Probability = pf;
          Params->Symmetry |= REFLEXION; 
          break;
  case 5:
          pf = GeneralizedCauchyEV;
          Params->Probability = pf;
          Params->Symmetry |= REFLEXION;
          break;
  case 6:
          pf = GeneralizedLogistic;
          Params->Probability = pf;
          break;
  case 7:
          pf = GeneralizedLogisticEV;
          Params->Probability = pf;
          break;
  case 8:
          pf = GaussianTEV;
          Params->Probability = pf;
          Params->Symmetry |= (TRANSLATION | REFLEXION);
          break;
  default:
          pf = GaussianT;
          Params->Probability = pf;
          Params->Symmetry |= (TRANSLATION | REFLEXION);
  }
#else
  switch(probID)  {
  case 1:
          pf = GaussianT;
          Params->Probability = pf;
          Params->Symmetry |= (TRANSLATION | REFLEXION);
          break;
  case 2:
          pf = Gaussian;
          Params->Probability = pf;
          Params->Symmetry |= REFLEXION ;
          break;
  case 3:
          pf = Student3T;
          Params->Probability = pf;
          Params->Symmetry |= (TRANSLATION | REFLEXION) ;
          break;
  default:
          pf = GaussianT;
          Params->Probability = pf;
          Params->Symmetry |= (TRANSLATION | REFLEXION);
  }
#endif

}

void PROCESS_Free
                ( 
                PROCESS *ProcessParams, 
                int NumRates
                )
{
  int i;
  int err = 0;

  if(ProcessParams==NULL) {
        fprintf(stderr,"PROCESS_Free():cannot free null buffer for ProcessParams\n");
        exit(1);
  }
  if(ProcessParams->Probability == NULL) {
        fprintf(stderr,"PROCESS_Free():cannot free null ProbabilityFunction\n");
        exit(1);
  } else {
        free((void *) ProcessParams->Probability);
  }
  if(ProcessParams->Variance== NULL) {
        fprintf(stderr,"PROCESS_Free():cannot free null variance buffer\n");
        exit(1);
  } else {
        V_Free(ProcessParams->Variance);
  }

  if(ProcessParams->RateCouplings== NULL) {
        fprintf(stderr,"PROCESS_Free():cannot free null buffer for couplings\n");
        exit(1);
  } else {
        M_Free((void **) ProcessParams->RateCouplings, NumRates);
  }

  free((void *) ProcessParams);
}

